use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007003
use Test::Spelling 0.12;
use Pod::Wordlist;

set_spell_cmd('aspell list -l en -p ./xt/aspell-en.pws');
add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
Bugger
Dist
Manifest
Manual
Plugin
Read
Van
Zilla
de
lib
van
